@@ -5,7 +5,7 @@
   public domain.  Send questions/comments/complaints/performance data
   to dl@cs.oswego.edu
 
-* VERSION 2.6.2  Mon Jan  8 10:28:33 1996  Doug Lea  (dl at gee)
+* VERSION 2.6.3f  Sun Jun  2 17:39:38 1996  Doug Lea  (dl at gee)
   
    Note: There may be an updated version of this malloc obtainable at
            ftp://g.oswego.edu/pub/misc/malloc.c
@@ -100,6 +100,15 @@
             mmap(), the worst case wastage is 8 bytes plus the remainder
             from a system page (the minimal mmap unit); typically 4096 bytes.
 
+* Limitations
+
+    Here are some features that are NOT currently supported
+
+    * No support for 64-bit pointers.
+    * No user-definable hooks for callbacks and the like.
+    * No automated mechanism for fully checking that all accesses
+      to malloced memory stay within their bounds.
+    * No support for compaction.
 
 * Synopsis of compile-time options:
 
@@ -133,6 +142,9 @@
   HAVE_MMAP                 (default: defined as 1)
      Define to non-zero to optionally make malloc() use mmap() to
      allocate very large blocks.  
+  HAVE_MREMAP                 (default: defined as 0 unless Linux libc set)
+     Define to non-zero to optionally make realloc() use mremap() to
+     reallocate very large blocks.  
   malloc_getpagesize        (default: derived from system #includes)
      Either a constant or routine call returning the system page size.
   HAVE_USR_INCLUDE_MALLOC_H (default: NOT defined) 
@@ -149,6 +161,9 @@
      The name of the routine to call to obtain more memory from the system.
   MORECORE_FAILURE          (default: -1)
      The value returned upon failure of MORECORE.
+  MORECORE_CLEARS           (default 1)
+     True (1) if the routine mapped to MORECORE zeroes out memory (which
+     holds for sbrk).
   DEFAULT_TRIM_THRESHOLD
   DEFAULT_TOP_PAD       
   DEFAULT_MMAP_THRESHOLD
@@ -303,7 +318,11 @@ void* memcpy(void*, const void*, size_t);
 */
 
 #ifndef HAVE_MREMAP
+#ifdef INTERNAL_LINUX_C_LIB
 #define HAVE_MREMAP 1
+#else
+#define HAVE_MREMAP 0
+#endif
 #endif
 
 #if HAVE_MMAP
@@ -612,6 +631,7 @@ Void_t *(*__morecore)() = __default_morecore_init;
 
 #define MORECORE (*__morecore)
 #define MORECORE_FAILURE 0
+#define MORECORE_CLEARS 1 
 
 #else /* INTERNAL_LINUX_C_LIB */
 
@@ -629,6 +649,10 @@ extern Void_t*     sbrk();
 #define MORECORE_FAILURE -1
 #endif
 
+#ifndef MORECORE_CLEARS
+#define MORECORE_CLEARS 1
+#endif
+
 #endif /* INTERNAL_LINUX_C_LIB */
 
 #if defined(INTERNAL_LINUX_C_LIB) && defined(__ELF__)
@@ -716,7 +740,7 @@ struct malloc_chunk
   size_t size;               /* Size in bytes, including overhead. */
   struct malloc_chunk* fd;   /* double links -- used only if free. */
   struct malloc_chunk* bk;
-  size_t unused;             /* to pad decl to min chunk size */
+  size_t unused;             /* fake decl to pad out to min chunk size */
 };
 
 typedef struct malloc_chunk* mchunkptr;
@@ -1014,6 +1038,7 @@ typedef struct malloc_chunk* mbinptr;
 
 #define bin_at(i)      ((mbinptr)(&(av_[2 * (i)])))
 #define next_bin(b)    ((mbinptr)((char*)(b) + 2 * sizeof(mbinptr)))
+#define prev_bin(b)    ((mbinptr)((char*)(b) - 2 * sizeof(mbinptr)))
 
 /*
    The first 2 bins are never indexed. The corresponding av_ cells are instead
@@ -1245,7 +1270,10 @@ static void do_check_inuse_chunk(p) mchunkptr p;
     do_check_free_chunk(prv);
   }
   if (next == top)
+  {
     assert(prev_inuse(next));
+    assert(chunksize(next) >= MINSIZE);
+  }
   else if (!inuse(next))
     do_check_free_chunk(next);
 
@@ -1576,15 +1604,24 @@ static void malloc_extend_top(nb) size_t nb;
 
     if (old_top != initial_top)
     {
+
       /* There must have been an intervening foreign sbrk call. */
       /* A double fencepost is necessary to prevent consolidation */
+
+      /* If not enough space to do this, then user did something very wrong */
+      if (old_top_size < MINSIZE) 
+      {
+        set_head(top, PREV_INUSE); /* will force null return from malloc */
+        return;
+      }
+
       chunk_at_offset(old_top, old_top_size - 2*SIZE_SZ)->size = 
         SIZE_SZ|PREV_INUSE;
       chunk_at_offset(old_top, old_top_size -   SIZE_SZ)->size = 
         SIZE_SZ|PREV_INUSE;
 
       /* Also keep size a multiple of MINSIZE */
-      old_top_size = (old_top_size - 2*SIZE_SZ) & MALLOC_ALIGN_MASK;
+      old_top_size = (old_top_size - 2*SIZE_SZ) & ~MALLOC_ALIGN_MASK;
       chunk_at_offset(old_top, old_top_size          )->size =
         SIZE_SZ|PREV_INUSE;
       chunk_at_offset(old_top, old_top_size + SIZE_SZ)->size =
@@ -1687,6 +1724,7 @@ Void_t* mALLOc(bytes) size_t bytes;
   int       startidx;                /* first bin of a traversed block */
   mchunkptr fwd;                     /* misc temp for linking */
   mchunkptr bck;                     /* misc temp for linking */
+  mbinptr q;                         /* misc temp */
 
   size_t nb  = request2size(bytes);  /* padded request size; */
 
@@ -1697,10 +1735,17 @@ Void_t* mALLOc(bytes) size_t bytes;
     idx = smallbin_index(nb); 
 
     /* No traversal or size check necessary for small bins.  */
-    /* Also scan the next one, since it would have a remainder < MINSIZE */
 
-    if ( ((victim = last(bin_at(idx)))   != bin_at(idx)) || 
-         ((victim = last(bin_at(idx+1))) != bin_at(idx+1)))
+    q = bin_at(idx);
+    victim = last(q);
+
+    /* Also scan the next one, since it would have a remainder < MINSIZE */
+    if (victim == q)
+    {
+      q = next_bin(q);
+      victim = last(q);
+    }
+    if (victim != q)
     {
       victim_size = chunksize(victim);
       unlink(victim, bck, fwd);
@@ -1799,12 +1844,11 @@ Void_t* mALLOc(bytes) size_t bytes;
     for (;;)  
     {
       startidx = idx;          /* (track incomplete blocks) */
+      q = bin = bin_at(idx);
 
       /* For each bin in this block ... */
-      do           
-      { 
-        bin = bin_at(idx);
-
+      do
+      {
         /* Find and use first big enough chunk ... */
 
         for (victim = last(bin); victim != bin; victim = victim->bk)
@@ -1834,6 +1878,8 @@ Void_t* mALLOc(bytes) size_t bytes;
 
         }
 
+       bin = next_bin(bin);
+
       } while ((++idx & (BINBLOCKWIDTH - 1)) != 0);
 
       /* Clear out the block bit. */
@@ -1846,7 +1892,8 @@ Void_t* mALLOc(bytes) size_t bytes;
           break;
         }
         --startidx;
-      } while (first(bin_at(startidx)) == bin_at(startidx));
+       q = prev_bin(q);
+      } while (first(q) == q);
 
       /* Get to the next possibly nonempty block */
 
@@ -1932,7 +1979,7 @@ void fREe(mem) Void_t* mem;
   size_t    prevsz;    /* size of previous contiguous chunk */
   mchunkptr bck;       /* misc temp for linking */
   mchunkptr fwd;       /* misc temp for linking */
-  int       islr;      /* track whther merging with last_remainder */
+  int       islr;      /* track whether merging with last_remainder */
 
   if (mem == 0)                              /* free(0) has no effect */
     return;
@@ -2017,7 +2064,12 @@ void fREe(mem) Void_t* mem;
 
   Realloc algorithm:
 
-    If the reallocation is for additional space, and the chunk can be
+    Chunks that were obtained via mmap cannot be extended or shrunk
+    unless HAVE_MREMAP is defined, in wchi case mrep is used.
+    Otherwise, if their reallocation is for additional space, they are
+    copied.  If for less, they are just left alone.
+
+    Otherwise, if the reallocation is for additional space, and the chunk can be
     extended, it is, else a malloc-copy-free sequence is taken.  There
     are several different ways that a chunk could be extended. All are
     tried:
@@ -2027,15 +2079,12 @@ void fREe(mem) Void_t* mem;
        * Both shifting backwards and extending forward.
        * Extending into newly sbrked space
 
+    Unless the #define REALLOC_ZERO_BYTES_FREES is set, realloc with a
+    size argument of zero (re)allocates a minimum-sized chunk.
 
-    If the reallocation is for less space, the trailing space is
-    lopped off and freed.  Unless the #define REALLOC_ZERO_BYTES_FREES
-    is set, realloc with a size argument of zero (re)allocates a
-    minimum-sized chunk.
-
-    Chunks that were obtained via mmap cannot be extended or shrunk.
-    If their reallocation is for additional space, they are copied.
-    If for less, they are just left alone.
+    If the reallocation is for less space, and the new request is for
+    a `small' (<512 bytes) size, then the newly unused space is lopped
+    off and freed.
 
     The old unix realloc convention of allowing the last-free'd chunk
     to be used as an argument to realloc is no longer supported.
@@ -2220,6 +2269,7 @@ Void_t* rEALLOc(oldmem, bytes) Void_t* oldmem; size_t bytes;
     return newmem;
   }
 
+
  split:  /* split off extra room in old or expanded chunk */
 
   if (newsize - nb >= MINSIZE) /* split off remainder */
@@ -2337,6 +2387,8 @@ Void_t* mEMALIGn(alignment, bytes) size_t alignment; size_t bytes;
     set_head_size(p, leadsize);
     fREe(chunk2mem(p));
     p = newp;
+
+    assert (newsize >= nb && (((size_t)(chunk2mem(p))) % alignment) == 0);
   }
 
   /* Also give back spare room at the end */
@@ -2391,6 +2443,12 @@ Void_t* cALLOc(n, elem_size) size_t n; size_t elem_size;
   size_t csz;
 
   size_t sz = n * elem_size;
+
+  /* check if expand_top called, in which case don't need to clear */
+#if MORECORE_CLEARS
+  mchunkptr oldtop = top;
+  size_t oldtopsize = chunksize(top);
+#endif
   Void_t* mem = mALLOc (sz);
 
   if (mem == 0) 
@@ -2399,10 +2457,24 @@ Void_t* cALLOc(n, elem_size) size_t n; size_t elem_size;
   {
     p = mem2chunk(mem);
 
-    if(chunk_is_mmapped(p)) /* no clearing is necessary */
-      return mem;
+    /* Two optional cases in which clearing not necessary */
+
+
+#if HAVE_MMAP
+    if (chunk_is_mmapped(p)) return mem;
+#endif
 
     csz = chunksize(p);
+
+#if MORECORE_CLEARS
+  if (p == oldtop && csz > oldtopsize) 
+  {
+    /* clear only the bytes from non-freshly-sbrked memory */
+    MALLOC_ZERO(mem, oldtopsize - SIZE_SZ);
+    return mem;
+  }
+#endif
+
     MALLOC_ZERO(mem, csz - SIZE_SZ);
     return mem;
   }
@@ -2672,6 +2744,12 @@ int mALLOPt(param_number, value) int param_number; int value;
 
 History:
 
+    V2.6.3 Sun May 19 08:17:58 1996  Doug Lea  (dl at gee)
+      * Malloc, calloc: add optimizations from Raymond Nijssen
+      * malloc_extend_top: fix mask error that caused wastage after
+        foreign sbrks
+      * Add linux mremap support code from HJ Liu
+   
     V2.6.2 Tue Dec  5 06:52:55 1995  Doug Lea  (dl at gee)
       * Integrated most documentation with the code.
       * Add support for mmap, with help from 
