@@ -174,6 +174,10 @@ res_send_setrhook(hook)
 	Rhook = hook;
 }
 
+#ifdef __ELF__
+#pragma weak __res_isourserver = res_isourserver
+#endif
+
 /* int
  * res_isourserver(ina)
  *	looks up "ina" in _res.ns_addr_list[]
@@ -206,6 +210,10 @@ res_isourserver(inp)
 	return (ret);
 }
 
+#ifdef __ELF__
+#pragma weak __res_nameinquery = res_nameinquery
+#endif
+
 /* int
  * res_nameinquery(name, type, class, buf, eom)
  *	look for (name,type,class) in the query section of packet (buf,eom)
@@ -243,6 +251,10 @@ res_nameinquery(name, type, class, buf, eom)
 	return (0);
 }
 
+#ifdef __ELF__
+#pragma weak __res_queriesmatch = res_queriesmatch
+#endif
+
 /* int
  * res_queriesmatch(buf1, eom1, buf2, eom2)
  *	is there a 1:1 mapping of (name,type,class)
@@ -587,6 +599,11 @@ res_send(buf, buflen, ans, anssiz)
 			n = select(s+1, &dsmask, (fd_set *)NULL,
 				   (fd_set *)NULL, &timeout);
 			if (n < 0) {
+				/* This is for those applications who
+				 * sets to signal hanlder but not
+				 * restarting system calls. */
+				if (errno == EINTR)
+				  goto wait;
 				Perror(stderr, "select", errno);
 				_res_close();
 				goto next_ns;
