@@ -180,7 +180,7 @@ static struct toktab {
 static void
 rnetrc(const char *host, char **aname, char **apass)
 {
-	char *hdir, buf[BUFSIZ];
+	char *hdir, *buf;
 	int t;
 	struct stat stb;
 	struct hostent *hp;
@@ -188,9 +188,20 @@ rnetrc(const char *host, char **aname, char **apass)
 #if NLS
 	libc_nls_init();
 #endif
-	hdir = getenv("HOME");
-	if (hdir == NULL)
-		hdir = ".";
+	/*
+	 * Use __libc_secure_getenv -- otherwise you can probably read any
+	 * file on the system by setting $HOME to /tmp/foo and linking 
+	 * /tmp/foo/.netrc, if you can find a setuid program using this code.
+	 * Also, if we can't get HOME, fail instead of trying ".", which is
+	 * no improvement. This really should call getpwuid(getuid()).
+	 */
+	hdir = __libc_secure_getenv("HOME");
+	if (hdir == NULL) {
+		/* hdir = "."; */
+		/* NO! return instead */
+		return;
+	}
+	buf = alloca (strlen (hdir) + sizeof ("/.netrc") + 1);
 	(void)sprintf(buf, "%s/.netrc", hdir);
 	cfile = fopen(buf, "r");
 	if (cfile == NULL) {
@@ -351,6 +362,10 @@ static char	e[] = {
 };
 
 #if 0
+/*
+ * Warning -- do not enable any of this without checking it carefully
+ * for buffer overflows.
+ */
 static
 char *nbsencrypt(str,key,result)
   char *str, *key;
@@ -856,17 +871,20 @@ char *mkenvkey(char mch)
 {
 	static char skey[40];
 	register struct utmp *putmp;
-	char stemp[40], stemp1[40], sttyname[30];
+	char stemp[40], stemp1[40];
+	/* FUTURE: remove the limit on this size */
+	char sttyname[30];
 	register char *sk,*p;
 
 	if (isatty(2))
-		strcpy(sttyname,ttyname(2));
+		strncpy(sttyname,ttyname(2), sizeof(sttyname));
 	else if (isatty(0))
-		strcpy(sttyname,ttyname(0));
+		strncpy(sttyname,ttyname(0), sizeof(sttyname));
 	else if (isatty(1))
-		strcpy(sttyname,ttyname(1));
+		strncpy(sttyname,ttyname(1), sizeof(sttyname));
 	else
 		return (NULL);
+	sttyname[sizeof(sttyname)-1] = 0;
 	putmp = getutmp(sttyname);
 	if (putmp == NULL)
 		return (NULL);
