@@ -170,15 +170,34 @@ int type;
 		for (pathP = path; *nlspath; ++nlspath) {
 		    if (*nlspath == '%') {
 			if (*(nlspath + 1) == 'L') {
+			    char * tmp = pathP + strlen(lang);
+			    if (tmp > path + sizeof (path))
+			    {
+			      /* overflow */
+			      goto error;
+			    }
 			    ++nlspath;
 			    strcpy(pathP, lang);
-			    pathP += strlen(lang);
+			    pathP = tmp;
 			} else if (*(nlspath + 1) == 'N') {
+			    char * tmp = pathP + strlen(name);
+			    if (tmp > path + sizeof (path))
+			    {
+			      /* overflow */
+			      goto error;
+			    }
 			    ++nlspath;
 			    strcpy(pathP, name);
-			    pathP += strlen(name);
+			    pathP = tmp;
 			} else *(pathP++) = *nlspath;
-		    } else *(pathP++) = *nlspath;
+		    } else
+		    {
+			if (pathP >= path + sizeof (path))
+			{
+				goto error;
+			}
+			*(pathP++) = *nlspath;
+		    }
 		}
 		*pathP = '\0';
 		if (stat(path, &sbuf) == 0) {
@@ -188,6 +207,7 @@ int type;
 		nlspath = cptr+1;
 	    }
 	}
+error:
 	free(base);
 	if (tmppath) free(tmppath);
 
