@@ -81,6 +81,10 @@ extern char *	getenv();
 
 #endif /* !defined __STDC__ */
 
+#include <libc-lock.h>
+__libc_lock_define (extern, __libc_localtime_lock);
+__libc_lock_define (extern, __libc_gmtime_lock);
+
 extern time_t	time();
 
 #define ACCESS_MODE	O_RDONLY
@@ -211,7 +215,6 @@ time_t			timezone = 0;
 int			daylight = 0;
 #endif /* defined USG_COMPAT */
 
-#define ALTZONE
 #ifdef ALTZONE
 time_t			altzone = 0;
 #endif /* defined ALTZONE */
@@ -291,6 +294,53 @@ settzname()
 	}
 }
 
+static int
+openFile (const char *name, char *buf, int size)
+{
+  char *prev, *cur;
+  int len = strlen (name) + 2;
+  int fd, l;
+
+  for (prev = TZDIR; cur = strchr (prev, ':'); prev = cur + 1)
+  {
+    l = cur - prev;
+
+    /* Check the buffer size */
+    if (l + len > size)
+    {
+      return -1;
+    }
+
+    strncpy (buf, prev, l);
+    buf [l++] = '/';
+    strcpy (&buf [l], name);
+
+    fd = __open (buf, OPEN_MODE);
+    if (fd >= 0 || errno != ENOENT)
+    {
+      /* Return if we open the file or error other than the file
+       * doesn't exist.
+       */
+      return fd;
+    }
+  }
+
+  l = strlen (prev);
+  /* Check the buffer size */
+  if (l + len > size)
+  {
+    return -1;
+  }
+
+  strcpy (buf, prev);
+  buf [l++] = '/';
+  strcpy (&buf [l], name);
+
+  fd = __open (buf, OPEN_MODE);
+
+  return fd;
+}
+
 static int
 tzload(name, sp)
 register const char *		name;
@@ -310,14 +360,14 @@ register struct state * const	sp;
 		if (name[0] != '/') {
 			if ((p = TZDIR) == NULL)
 				return -1;
-			if ((strlen(p) + strlen(name) + 1) >= sizeof fullname)
-				return -1;
-			(void) strcpy(fullname, p);
-			(void) strcat(fullname, "/");
-			(void) strcat(fullname, name);
-			name = fullname;
+			fid = openFile (name, fullname,
+				sizeof (fullname));
 		}
-		if ((fid = __open(name, OPEN_MODE)) == -1)
+		else
+		{
+			fid = __open(name, OPEN_MODE);
+		}
+		if (fid < 0)
 			return -1;
 	}
 	{
@@ -965,11 +1015,13 @@ struct tm * const	tmp;
 	register int			i;
 	const time_t			t = *timep;
 
+	 __libc_lock_lock(__libc_localtime_lock);
 	if (!lcl_is_set)
 		tzset();
 	sp = lclptr;
 #ifdef ALL_STATE
 	if (sp == NULL) {
+		__libc_lock_unlock(__libc_localtime_lock);
 		gmtsub(timep, offset, tmp);
 		return;
 	}
@@ -1000,6 +1052,7 @@ struct tm * const	tmp;
 #ifdef HAVE_TM_ZONE
 	tmp->tm_zone = &sp->chars[ttisp->tt_abbrind];
 #endif
+	__libc_lock_unlock(__libc_localtime_lock);
 }
 
 struct tm *
@@ -1012,6 +1065,16 @@ const time_t * const	timep;
 	return &tm;
 }
 
+struct tm *
+localtime_r (timep, tm)
+const time_t * const	timep;
+struct tm *tm;
+{
+
+	localsub(timep, 0L, tm);
+	return tm;
+}
+
 /*
 ** gmtsub is to gmtime as localsub is to localtime.
 */
@@ -1022,6 +1085,7 @@ const time_t * const	timep;
 const long		offset;
 struct tm * const	tmp;
 {
+	__libc_lock_lock(__libc_gmtime_lock);
 	if (!gmt_is_set) {
 		gmt_is_set = TRUE;
 #ifdef ALL_STATE
@@ -1030,6 +1094,7 @@ struct tm * const	tmp;
 #endif /* defined ALL_STATE */
 			gmtload(gmtptr);
 	}
+	__libc_lock_unlock(__libc_gmtime_lock);
 	timesub(timep, offset, gmtptr, tmp);
 	/*
 	** Could get fancy here and deliver something such as
@@ -1052,6 +1117,16 @@ struct tm * const	tmp;
 #endif /* HAVE_TM_ZONE */
 }
 
+struct tm *
+gmtime_r (timep, tm)
+const time_t * const	timep;
+struct tm *tm;
+
+{
+	gmtsub(timep, 0L, tm);
+	return tm;
+}
+
 struct tm *
 gmtime(timep)
 const time_t * const	timep;
@@ -1161,8 +1236,9 @@ register struct tm * const		tmp;
 */
 
 char *
-asctime(timeptr)
+asctime_r(timeptr, result)
 register const struct tm *	timeptr;
+char	*result;
 {
 	static const char	wday_name[DAYSPERWEEK][3] = {
 		"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
@@ -1171,7 +1247,6 @@ register const struct tm *	timeptr;
 		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
 		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
 	};
-	static char	result[26];
 
 	(void) sprintf(result, "%.3s %.3s%3d %02.2d:%02.2d:%02.2d %d\n",
 		wday_name[timeptr->tm_wday],
@@ -1182,6 +1257,14 @@ register const struct tm *	timeptr;
 	return result;
 }
 
+char *
+asctime(timeptr)
+register const struct tm *	timeptr;
+{
+	static char	result[26];
+	return asctime_r (timeptr, result);
+}
+
 char *
 ctime(timep)
 const time_t * const	timep;
@@ -1189,6 +1272,15 @@ const time_t * const	timep;
 	return asctime(localtime(timep));
 }
 
+char *
+ctime_r(timep, result)
+const time_t * const	timep;
+char *result;
+{
+	struct tm tm;
+	return asctime_r (localtime_r (timep, &tm), result);
+}
+
 /*
 ** Adapted from code provided by Robert Elz, who writes:
 **	The "best" way to do mktime I think is based on an idea of Bob
@@ -1212,13 +1304,12 @@ const int	base;
 		*tensptr += *unitsptr / base;
 		*unitsptr %= base;
 	} else if (*unitsptr < 0) {
-		--*tensptr;
-		*unitsptr += base;
-		if (*unitsptr < 0) {
-			*tensptr -= 1 + (-*unitsptr) / base;
-			*unitsptr = base - (-*unitsptr) % base;
-		}
+		int del;
+		del = (base - 1 - *unitsptr) / base;
+		*unitsptr += del*base;
+		*tensptr -= del;
 	}
+			/* assert (0 <= *unitsptr && *unitsptr < base); */
 }
 
 static int
@@ -1263,7 +1354,9 @@ int * const		okayp;
 	while (yourtm.tm_mday <= 0) {
 		--yourtm.tm_year;
 		yourtm.tm_mday +=
-			year_lengths[isleap(yourtm.tm_year + TM_YEAR_BASE)];
+			year_lengths[isleap(yourtm.tm_year 
+					    + (yourtm.tm_mon > 1) 
+					    + TM_YEAR_BASE)];
 	}
 	for ( ; ; ) {
 		i = mon_lengths[isleap(yourtm.tm_year +
